/*
Copyright (c) 2021 William Foote

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

///
/// Memory-efficient version of ScalableImage, at the expense of rendering
/// time.  This representation rendered about 3x slower in some informal
/// tests, but occupies perhaps an order of magnitude less memory.
///
library jovial_svg.compact;

import 'dart:convert';
import 'dart:typed_data';
import 'dart:ui';

import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:jovial_misc/io_utils.dart';
import 'affine.dart';
import 'dag.dart';
import 'common.dart';
import 'common_noui.dart';
import 'compact_noui.dart';
import 'exported.dart';
import 'path_noui.dart';
import 'path.dart';

///
/// A Scalable Image that's represented by a compact packed binary format
/// that is interpreted when rendering.
///
class ScalableImageCompact extends ScalableImageBase
    with ScalableImageCompactGeneric<Color, BlendMode, SIImage> {
  @override
  final bool bigFloats;
  @override
  final int numPaths;
  @override
  final int numPaints;
  @override
  final List<String> strings;
  @override
  final List<List<double>> floatLists;
  @override
  final Uint8List children;
  @override
  final List<double> args; // Float32List or Float64List
  @override
  final List<double> transforms; // Float32List or Float64List

  ScalableImageCompact._p(
      {required this.bigFloats,
      required double? width,
      required double? height,
      required Color? tintColor,
      required BlendMode tintMode,
      required Color? currentColor,
      required this.numPaths,
      required this.numPaints,
      required List<SIImage> images,
      required this.strings,
      required this.floatLists,
      required this.children,
      required this.args,
      required this.transforms,
      required Rect? viewport})
      : super(
            width, height, tintColor, tintMode, viewport, images, currentColor);

  @override
  ScalableImageCompact withNewViewport(Rect viewport,
      {bool prune = false, double pruningTolerance = 0}) {
    if (prune) {
      final v = _PruningVisitor(this, viewport, pruningTolerance);
      accept(v);
      return v.si;
    } else {
      return ScalableImageCompact._p(
          bigFloats: bigFloats,
          width: width,
          height: height,
          tintColor: tintColor,
          tintMode: tintMode,
          currentColor: currentColor,
          numPaths: numPaths,
          numPaints: numPaints,
          strings: strings,
          floatLists: floatLists,
          images: images,
          children: children,
          args: args,
          transforms: transforms,
          viewport: viewport);
    }
  }

  @override
  ScalableImage modifyCurrentColor(Color newCurrentColor) {
    return ScalableImageCompact._p(
        bigFloats: bigFloats,
        width: width,
        height: height,
        tintColor: tintColor,
        tintMode: tintMode,
        currentColor: newCurrentColor,
        numPaths: numPaths,
        numPaints: numPaints,
        strings: strings,
        floatLists: floatLists,
        images: images,
        children: children,
        args: args,
        transforms: transforms,
        viewport: viewport);
  }

  @override
  ScalableImage modifyTint(
      {required BlendMode newTintMode, required Color? newTintColor}) {
    return ScalableImageCompact._p(
        bigFloats: bigFloats,
        width: width,
        height: height,
        tintColor: newTintColor,
        tintMode: newTintMode,
        currentColor: currentColor,
        numPaths: numPaths,
        numPaints: numPaints,
        strings: strings,
        floatLists: floatLists,
        images: images,
        children: children,
        args: args,
        transforms: transforms,
        viewport: viewport);
  }

  @override
  void paintChildren(Canvas c, Color currentColor) =>
      accept(_PaintingVisitor(c, RenderContext.root(this, currentColor)));

  @override
  PruningBoundary? getBoundary() => accept(_BoundaryVisitor(this));

  R accept<R>(SIVisitor<CompactChildData, SIImage, R> visitor) {
    final t = CompactTraverser<R, SIImage>(
        bigFloats: bigFloats,
        strings: strings,
        floatLists: floatLists,
        images: images,
        visiteeChildren: children,
        visiteeArgs: args,
        visiteeTransforms: transforms,
        visiteeNumPaths: numPaths,
        visiteeNumPaints: numPaints,
        visitor: visitor);
    R r = t.traverse(visitor.initial);
    return r;
  }

  @override
  ScalableImageDag toDag() {
    final b = SIDagBuilderFromCompact(viewport,
        warn: false, currentColor: currentColor);
    b.vector(
        width: width,
        height: height,
        tintColor: tintColor?.value,
        tintMode: SITintModeMapping.fromBlendMode(tintMode));
    accept(b);
    b.endVector();
    return b.si;
  }

  @override
  SITintMode blendModeToSI(BlendMode b) => SITintModeMapping.fromBlendMode(b);

  @override
  int colorValue(Color tintColor) => tintColor.value;

  static ScalableImageCompact fromByteData(ByteData data,
          {Color? currentColor}) =>
      fromBytes(
          Uint8List.view(data.buffer, data.offsetInBytes, data.lengthInBytes),
          currentColor: currentColor);

  static ScalableImageCompact fromBytes(Uint8List data, {Color? currentColor}) {
    final dis = ByteBufferDataInputStream(data, Endian.big);
    final magic = dis.readUnsignedInt();
    if (magic != ScalableImageCompactGeneric.magicNumber) {
      throw ParseError('Bad magic number:  0x${magic.toRadixString(16)}');
    }
    dis.readByte();
    final version = dis.readUnsignedShort();
    if (version != ScalableImageCompactGeneric.fileVersionNumber) {
      throw ParseError('Unsupported version $version');
    }
    final int flags = dis.readUnsignedByte();
    final hasWidth = _flag(flags, 0);
    final hasHeight = _flag(flags, 1);
    final bigFloats = _flag(flags, 2);
    final hasTintColor = _flag(flags, 3);
    final numPaths = dis.readUnsignedInt();
    final numPaints = dis.readUnsignedInt();
    final argsLen = dis.readUnsignedInt();
    final xformsLen = dis.readUnsignedInt();
    final List<double> args = _floatList(dis, bigFloats, argsLen);
    final List<double> transforms = _floatList(dis, bigFloats, xformsLen);
    final width = _readFloat(dis, bigFloats, hasWidth);
    final height = _readFloat(dis, bigFloats, hasHeight);
    final int? tintColor;
    final int tintMode;
    if (hasTintColor) {
      tintColor = dis.readUnsignedInt();
      tintMode = dis.readUnsignedByte();
    } else {
      tintMode = SITintModeMapping.defaultValue.index;
      tintColor = null;
    }
    final strings = List<String>.generate(
        _readSmallishInt(dis),
        (_) => (const Utf8Decoder(allowMalformed: true))
            .convert(dis.readBytesImmutable(_readSmallishInt(dis))),
        growable: false);
    final floatLists = List<List<double>>.generate(_readSmallishInt(dis),
        (_) => _floatList(dis, bigFloats, _readSmallishInt(dis), Endian.big),
        growable: false);
    final images = List<SIImageData>.generate(_readSmallishInt(dis), (_) {
      final x = _readFloat(dis, bigFloats, true)!;
      final y = _readFloat(dis, bigFloats, true)!;
      final w = _readFloat(dis, bigFloats, true)!;
      final h = _readFloat(dis, bigFloats, true)!;
      final encoded = dis.readBytes(_readSmallishInt(dis));
      return SIImageData(x: x, y: y, width: w, height: h, encoded: encoded);
    }, growable: false);
    final children = dis.remainingCopy();
    return ScalableImageCompact._p(
        bigFloats: bigFloats,
        width: width,
        height: height,
        tintColor: tintColor == null ? null : Color(tintColor),
        tintMode: SITintMode.values[tintMode].asBlendMode,
        currentColor: currentColor,
        numPaths: numPaths,
        numPaints: numPaints,
        strings: strings,
        floatLists: floatLists,
        images: List<SIImage>.generate(images.length, (i) => SIImage(images[i]),
            growable: false),
        children: children,
        args: args,
        transforms: transforms,
        viewport: null);
  }

  @override
  SIImageData getImageData(SIImage image) => image.data;

  static bool _flag(int byte, int bitNumber) => (byte & (1 << bitNumber)) != 0;

  static List<double> _floatList(
      ByteBufferDataInputStream dis, bool bigFloats, int length,
      [Endian endian = Endian.little]) {
    if (bigFloats) {
      Uint8List bytes = dis.readBytes(length * 8);
      ByteData bd = bytes.buffer.asByteData(bytes.offsetInBytes, length * 8);
      final r = Float64List(length);
      for (int i = 0; i < length; i++) {
        r[i] = bd.getFloat64(8 * i, endian);
      }
      return r;
    } else {
      Uint8List bytes = dis.readBytes(length * 4);
      ByteData bd = bytes.buffer.asByteData(bytes.offsetInBytes, length * 4);
      final r = Float32List(length);
      for (int i = 0; i < length; i++) {
        r[i] = bd.getFloat32(4 * i, endian);
      }
      return r;
    }
  }

  static double? _readFloat(
      ByteBufferDataInputStream dis, bool bigFloats, bool notNull) {
    if (notNull) {
      if (bigFloats) {
        return dis.readDouble();
      } else {
        return dis.readFloat();
      }
    } else {
      return null;
    }
  }

  static int _readSmallishInt(ByteBufferDataInputStream str) =>
      ScalableImageCompactGeneric.readSmallishInt(str);

  @override
  Uint8List toSIBytes() {
    final sink = ByteSink();
    final os = DataOutputSink(sink, Endian.big);
    writeToFile(os);
    os.close();
    return sink.toList();
  }
}

///
/// Helper for visitors of compact scalable images.  This class adds
/// the creation of renderable SI objects.
///
abstract class _CompactVisitor<R>
    with SIGroupHelper
    implements SIVisitor<CompactChildData, SIImage, R> {
  late final List<String> _strings;
  late final List<List<double>> _floatLists;
  late final List<SIImage> _images;
  RenderContext _context;
  RenderContext get context => _context;

  _CompactVisitor(this._context);

  @override
  R init(R collector, List<SIImage> images, List<String> strings,
      List<List<double>> floatLists) {
    _images = images;
    _strings = strings;
    _floatLists = floatLists;
    return collector;
  }

  @override
  R path(R collector, CompactChildData pathData, SIPaint paint) {
    final pb = UIPathBuilder();
    CompactPathParser(pathData, pb).parse();
    final p = SIPath(pb.path, paint);
    return siPath(collector, p);
  }

  R siPath(R collector, SIPath p);

  @override
  R clipPath(R collector, CompactChildData pathData) {
    final pb = UIPathBuilder();
    CompactPathParser(pathData, pb).parse();
    final p = SIClipPath(pb.path);
    return siClipPath(collector, p);
  }

  R siClipPath(R collector, SIClipPath path);

  @override
  R text(R collector, int xIndex, int yIndex, int textIndex,
      SITextAttributes ta, int? fontFamilyIndex, SIPaint p) {
    return siText(
        collector,
        SIText(_strings[textIndex], _floatLists[xIndex], _floatLists[yIndex],
            ta, p),
        xIndex,
        yIndex);
  }

  R siText(R collector, SIText text, int xIndex, int yIndex);

  void pushContext(RenderContext c) {
    assert(c.parent == _context);
    _context = c;
  }

  void popContext() {
    final c = _context.parent;
    if (c == null) {
      throw StateError('More pops than pushes');
    }
    _context = c;
  }

  @override
  void assertDone() {
    assert(_context.parent == null);
  }
}

class _PaintingVisitor extends _CompactVisitor<void> {
  final Canvas canvas;

  _PaintingVisitor(this.canvas, RenderContext context) : super(context);

  @override
  void get initial {}

  @override
  void group(void collector, Affine? transform, int? groupAlpha) {
    startPaintGroup(canvas, transform, groupAlpha);
    pushContext(RenderContext(context, transform: transform));
  }

  @override
  void endGroup(void collector) {
    endPaintGroup(canvas);
    popContext();
  }

  @override
  void siPath(void collector, SIPath path) => path.paint(canvas, context);

  @override
  void siClipPath(void collector, SIClipPath path) =>
      path.paint(canvas, context);

  @override
  void image(void collector, int imageIndex) =>
      _images[imageIndex].paint(canvas, context);

  @override
  void siText(void collector, SIText text, int xIndex, int yIndex) =>
      text.paint(canvas, context);
}

class _PruningVisitor extends _CompactVisitor<PruningBoundary> {
  final PruningBoundary _boundary;
  final _groupStack = List<_PruningEntry>.empty(growable: true);
  final _PruningBuilder builder;
  ScalableImageCompact? _si;
  CompactChildData? _lastPathData;
  final _theCanon = CanonicalizedData<SIImage>();

  _PruningVisitor(ScalableImageCompact si, Rect viewport, double tolerance)
      : _boundary = PruningBoundary(viewport.deflate(tolerance)),
        builder = _PruningBuilder(
            si.bigFloats,
            ByteSink(),
            (si.bigFloats) ? Float64Sink() : Float32Sink(),
            (si.bigFloats) ? Float64Sink() : Float32Sink(),
            viewport,
            currentColor: si.currentColor,
            warn: false),
        super(RenderContext.root(si, Colors.black)) {
    builder.vector(
        width: viewport.width,
        height: viewport.height,
        tintColor: si.tintColor?.value,
        tintMode: SITintModeMapping.fromBlendMode(si.tintMode));
  }

  @override
  PruningBoundary get initial => _boundary;

  ScalableImageCompact get si {
    final r = _si;
    if (r != null) {
      return r;
    } else {
      builder.endVector();
      builder.setCanon(_theCanon);
      return _si = builder.si;
    }
  }

  @override
  PruningBoundary group(
      PruningBoundary boundary, Affine? transform, int? groupAlpha) {
    final parent = _groupStack.isEmpty ? null : _groupStack.last;
    _groupStack
        .add(_PruningEntry(boundary, parent, this, transform, groupAlpha));
    pushContext(RenderContext(context, transform: transform));
    return context.transformBoundaryFromParent(boundary);
  }

  @override
  PruningBoundary endGroup(PruningBoundary boundary) {
    final us = _groupStack.last;
    _groupStack.length = _groupStack.length - 1;
    us.endGroupIfNeeded();
    popContext();
    return us.boundary;
  }

  @override
  PruningBoundary path(
      PruningBoundary boundary, CompactChildData pathData, SIPaint siPaint) {
    assert(_lastPathData == null);
    _lastPathData = CompactChildData.copy(pathData);
    return super.path(boundary, pathData, siPaint);
  }

  @override
  PruningBoundary siPath(PruningBoundary boundary, SIPath path) {
    assert(_lastPathData != null);
    if (path.prunedBy({}, {}, boundary) != null) {
      if (_groupStack.isNotEmpty) {
        _groupStack.last.generateGroupIfNeeded();
      }
      builder.path(null, _lastPathData!, path.siPaint);
    }
    _lastPathData = null;
    return boundary;
  }

  @override
  PruningBoundary clipPath(
      PruningBoundary boundary, CompactChildData pathData) {
    assert(_lastPathData == null);
    _lastPathData = CompactChildData.copy(pathData);
    return super.clipPath(boundary, pathData);
  }

  @override
  PruningBoundary siClipPath(PruningBoundary boundary, SIClipPath cp) {
    assert(_lastPathData != null);
    if (cp.prunedBy({}, {}, boundary) != null) {
      if (_groupStack.isNotEmpty) {
        _groupStack.last.generateGroupIfNeeded();
      }
      builder.clipPath(null, _lastPathData!);
    }
    _lastPathData = null;
    return boundary;
  }

  @override
  PruningBoundary image(PruningBoundary boundary, int imageIndex) {
    final SIImage image = _images[imageIndex];
    if (image.prunedBy({}, {}, boundary) != null) {
      if (_groupStack.isNotEmpty) {
        _groupStack.last.generateGroupIfNeeded();
      }
      int i = _theCanon.getIndex(_theCanon.images, image)!;
      builder.image(null, i);
    }
    return boundary;
  }

  @override
  PruningBoundary siText(
      PruningBoundary boundary, SIText text, int xIndex, int yIndex) {
    if (text.prunedBy({}, {}, boundary) != null) {
      if (_groupStack.isNotEmpty) {
        _groupStack.last.generateGroupIfNeeded();
      }
      final c = _theCanon;
      xIndex = c.getIndex(c.floatLists, _floatLists[xIndex])!;
      yIndex = c.getIndex(c.floatLists, _floatLists[yIndex])!;
      int textIndex = c.getIndex(c.strings, text.text)!;
      SITextAttributes ta = text.attributes;
      int fontFamilyIndex = c.getIndex(c.strings, ta.fontFamily)!;
      SIPaint p = text.siPaint;
      builder.text(null, xIndex, yIndex, textIndex, ta, fontFamilyIndex, p);
    }
    return boundary;
  }
}

class _PruningBuilder extends SIGenericCompactBuilder<CompactChildData, SIImage>
    with _SICompactPathBuilder {
  final Rect viewport;
  final Color? currentColor;

  @override
  void get initial {}

  _PruningBuilder(bool bigFloats, ByteSink childrenSink, FloatSink args,
      FloatSink transforms, this.viewport,
      {required bool warn, required this.currentColor})
      : super(bigFloats, childrenSink,
            DataOutputSink(childrenSink, Endian.little), args, transforms,
            warn: warn);

  @override
  void init(void collector, List<SIImage> images, List<String> strings,
      List<List<double>> floatLists) {
    // This is a little tricky.  When pruning, we collect the canonicalized
    // data on the fly, and only provide it to our supertype at the end,
    // right before building the SI, so we need to discard this data here, which
    // comes from the graph being pruned.
  }

  ///
  /// Set the canonicalized data.  This can be done right before
  /// calling `si`.
  ///
  void setCanon(CanonicalizedData<SIImage> canon) {
    super.init(null, canon.toList(canon.images), canon.toList(canon.strings),
        canon.toList(canon.floatLists));
  }

  ScalableImageCompact get si {
    assert(done);
    return ScalableImageCompact._p(
        bigFloats: bigFloats,
        width: width,
        height: height,
        tintColor: (tintColor == null) ? null : Color(tintColor!),
        tintMode: tintMode.asBlendMode,
        currentColor: currentColor,
        numPaths: numPaths,
        numPaints: numPaints,
        strings: strings,
        floatLists: floatLists,
        images: images,
        children: childrenSink.toList(),
        args: args.toList(),
        transforms: transforms.toList(),
        viewport: viewport);
  }
}

class _PruningEntry {
  final PruningBoundary boundary;
  final _PruningEntry? parent;
  final _PruningVisitor visitor;
  final Affine? transform;
  final int? groupAlpha;

  bool generated = false;

  _PruningEntry(this.boundary, this.parent, this.visitor, this.transform,
      this.groupAlpha);

  void generateGroupIfNeeded() {
    if (generated) {
      return;
    }
    parent?.generateGroupIfNeeded();
    visitor.builder.group(null, transform, groupAlpha);
    generated = true;
  }

  void endGroupIfNeeded() {
    if (generated) {
      visitor.builder.endGroup(null);
    }
  }
}

class _BoundaryVisitor extends _CompactVisitor<PruningBoundary?> {
  PruningBoundary? boundary;

  final _groupStack = List<PruningBoundary?>.empty(growable: true);

  _BoundaryVisitor(ScalableImageCompact si)
      : super(RenderContext.root(si, Colors.black));

  @override
  PruningBoundary? get initial => null;

  @override
  PruningBoundary? group(
      PruningBoundary? initial, Affine? transform, int? groupAlpha) {
    pushContext(RenderContext(context, transform: transform));
    _groupStack.add(initial);
    return null;
  }

  @override
  PruningBoundary? endGroup(PruningBoundary? children) {
    PruningBoundary? us = _groupStack.last;
    _groupStack.length = _groupStack.length - 1;
    RenderContext ctx = context;
    popContext();
    return combine(us, ctx.transformBoundaryFromChildren(children));
  }

  @override
  PruningBoundary? siPath(PruningBoundary? initial, SIPath path) {
    return combine(initial, path.getBoundary());
  }

  @override
  PruningBoundary? siClipPath(PruningBoundary? initial, SIClipPath cp) {
    return combine(initial, cp.getBoundary());
  }

  PruningBoundary? combine(PruningBoundary? a, PruningBoundary? b) {
    if (a == null) {
      return b;
    } else if (b == null) {
      return a;
    } else {
      return PruningBoundary(a.getBounds().expandToInclude(b.getBounds()));
      // See comment in _SIParentNode.getBoundary();
    }
  }

  @override
  PruningBoundary? image(PruningBoundary? collector, int imageIndex) {
    return combine(boundary, _images[imageIndex].getBoundary());
  }

  @override
  PruningBoundary? siText(
      PruningBoundary? boundary, SIText text, int xIndex, int yIndex) {
    return combine(boundary, text.getBoundary());
  }
}

mixin _SICompactPathBuilder {
  void makePath(CompactChildData pathData, PathBuilder pb, {bool warn = true}) {
    try {
      CompactPathParser(pathData, pb).parse();
    } catch (e) {
      if (warn) {
        print(e);
        // As per the SVG spec, paths shall be parsed up to the first error,
        // and it is recommended that errors be reported to the user if
        // posible.
      }
    }
  }

  CompactChildData immutableKey(CompactChildData key) =>
      CompactChildData.copy(key);
}

class SIDagBuilderFromCompact
    extends SIGenericDagBuilder<CompactChildData, SIImage>
    with _SICompactPathBuilder {
  SIDagBuilderFromCompact(Rect? viewport,
      {required bool warn, Color? currentColor})
      : super(viewport, warn, currentColor);

  @override
  List<SIImage> convertImages(List<SIImage> images) => images;
}

class SICompactBuilder extends SIGenericCompactBuilder<String, SIImageData>
    with SIStringPathMaker {
  final Color? currentColor;

  SICompactBuilder._p(bool bigFloats, ByteSink childrenSink, FloatSink args,
      FloatSink transforms, {required bool warn, this.currentColor})
      : super(bigFloats, childrenSink,
            DataOutputSink(childrenSink, Endian.little), args, transforms,
            warn: warn);

  @override
  void get initial {}

  factory SICompactBuilder(
      {required bool bigFloats, required bool warn, Color? currentColor}) {
    final cs = ByteSink();
    final a = bigFloats ? Float64Sink() : Float32Sink();
    final t = bigFloats ? Float64Sink() : Float32Sink();
    try {
      return SICompactBuilder._p(bigFloats, cs, a, t,
          warn: warn, currentColor: currentColor);
    } finally {
      cs.close();
    }
  }

  ScalableImageCompact get si {
    assert(done);
    return ScalableImageCompact._p(
        bigFloats: bigFloats,
        width: width,
        height: height,
        tintColor: (tintColor == null) ? null : Color(tintColor!),
        tintMode: tintMode.asBlendMode,
        currentColor: currentColor,
        numPaths: numPaths,
        numPaints: numPaints,
        strings: strings,
        floatLists: floatLists,
        images: List<SIImage>.generate(images.length, (i) => SIImage(images[i]),
            growable: false),
        children: childrenSink.toList(),
        args: args.toList(),
        transforms: transforms.toList(),
        viewport: null);
  }
}
